@@ -8,7 +8,7 @@ from typing import Union
 
 from celery import current_app
 from celery.result import AsyncResult
-from celery.signals import before_task_publish
+from celery.signals import before_task_publish, task_prerun
 from celery.states import FAILURE, REVOKED, PENDING, STARTED, RECEIVED, RETRY
 from celery.utils.log import get_task_logger
 from firexkit.broker import handle_broker_timeout
@@ -63,7 +63,7 @@ def get_result_logging_name(result: AsyncResult, name=None):
     return '%s[%s]' % (name, result)
 
 
-@before_task_publish.connect()
+@before_task_publish.connect
 def populate_task_info(sender, declare, headers, **_kwargs):
     task_info = {'name': sender}
     try:
@@ -74,6 +74,15 @@ def populate_task_info(sender, declare, headers, **_kwargs):
     current_app.backend.client.hmset(headers['id'], task_info)
 
 
+@task_prerun.connect
+def update_task_name(sender, task_id, *_args, **_kwargs):
+    # Although the name was populated in populate_task_info before_task_publish, the name
+    # can be inaccurate if it was a plugin. We can only over-write it with the accurate name
+    # at task_prerun.
+    task_info = {'name': sender.name}
+    current_app.backend.client.hmset(task_id, task_info)
+
+
 def mark_queues_ready(*queue_names: str):
     current_app.backend.client.sadd('QUEUES', *queue_names)
 
@@ -531,3 +540,31 @@ def first_non_chain_interrupted_exception(ex):
         e = e.__cause__
     return e
 
+
+def extract_and_filter(*args,
+                       extract_from_children: bool = True,
+                       extract_task_returns_only: bool = False,
+                       extract_from_parents: bool = False,
+                       **kwargs) -> Union[tuple, dict]:
+    """Wrapper for :func:`firexkit.result.get_results` and `firexkit.result.get_results_upto_parent`
+    that defaults `extract_from_children` to :const:`True` and `extract_task_returns_only` to :const:`False`
+
+    This wrapper only exists for legacy reasons since older chains might be relying on these old defaults for
+    `extract_from_children` and `extract_task_returns_only`.
+
+    Use :meth:`firexkit.result.get_results` or `firexkit.result.get_results_upto_parent` instead.
+
+    See Also:
+        firexkit.result.get_results
+    """
+
+    if extract_from_parents:
+        return get_results_upto_parent(*args,
+                                       return_keys_only=extract_task_returns_only,
+                                       merge_children_results=extract_from_children,
+                                       **kwargs)
+    else:
+        return get_results(*args,
+                           return_keys_only=extract_task_returns_only,
+                           merge_children_results=extract_from_children,
+                           **kwargs)
\ No newline at end of file
