@@ -6,8 +6,9 @@ import os
 import textwrap
 from collections import OrderedDict
 import inspect
-from typing import Callable, Iterable, Optional
+from typing import Callable, Iterable, Optional, Union
 from urllib.parse import urljoin
+from copy import deepcopy
 
 from celery.canvas import Signature
 from celery.result import AsyncResult
@@ -25,8 +26,9 @@ from firexkit.bag_of_goodies import BagOfGoodies
 from firexkit.argument_conversion import ConverterRegister
 from firexkit.result import get_tasks_names_from_results, wait_for_any_results, \
     RETURN_KEYS_KEY, wait_on_async_results_and_maybe_raise, get_result_logging_name, ChainInterruptedException, \
-    ChainRevokedException
-from firexkit.resources import get_firex_css_filepath, get_firex_logo_filepath, JINJA_ENV
+    ChainRevokedException, extract_and_filter
+from firexkit.resources import get_firex_css_filepath, get_firex_logo_filepath
+from firexkit.firexkit_common import JINJA_ENV
 import time
 
 logger = get_task_logger(__name__)
@@ -55,6 +57,106 @@ class DyanmicReturnsNotADict(Exception):
     pass
 
 
+def create_collapse_ops(flex_collapse_ops_spec):
+    from typing import Pattern
+
+    if isinstance(flex_collapse_ops_spec, list):
+        return flex_collapse_ops_spec
+
+    result_ops = []
+    if isinstance(flex_collapse_ops_spec, dict):
+        for k, v in flex_collapse_ops_spec.items():
+            op = {}
+            if isinstance(k, str):
+                op['relative_to_nodes'] = {'type': 'task_name', 'value': k}
+            elif issubclass(type(k), Pattern):  # can't use isinstance till 3.6
+                op['relative_to_nodes'] = {'type': 'task_name_regex', 'value': k.pattern}
+
+            if isinstance(v, str):
+                op['targets'] = [v]
+            elif isinstance(v, list):
+                op['targets'] = v
+            elif isinstance(v, dict):
+                # take targets, operation from value.
+                op.update(v)
+
+            if 'targets' in op:
+                result_ops.append(op)
+            else:
+                # TODO: fail or ignore malformed specs?
+                pass
+    elif isinstance(flex_collapse_ops_spec, str):
+        result_ops.append({'targets': [flex_collapse_ops_spec]})
+
+    return result_ops
+
+
+def expand_self_op():
+    return {'operation': 'expand', 'targets': ['self']}
+
+
+def flame_collapse(flex_collapse_ops):
+    def formatter(ops, task):
+        filled_ops = []
+        for op in ops:
+            if 'targets' not in op or not isinstance(op['targets'], list):
+                # Ignore malformed operations.
+                continue
+            filled_op = {
+                # defaults.
+                'relative_to_nodes': {'type': 'task_uuid', 'value': task.request.id},
+                'source_node': {'type': 'task_uuid', 'value': task.request.id},
+                'operation': 'collapse',
+            }
+            filled_op.update(op)
+            filled_ops.append(filled_op)
+        return filled_ops
+
+    static_ops = create_collapse_ops(flex_collapse_ops)
+    return flame('_default_display', formatter, data_type='object', bind=True, on_next=True, on_next_args=[static_ops],
+                 decorator_name=flame_collapse.__name__)
+
+
+def _default_flame_formatter(data):
+    if data is None:
+        return None
+    if not isinstance(data, str):
+        try:
+            return json.dumps(data)
+        except TypeError:
+            return str(data)
+    return data
+
+
+def create_flame_config(existing_configs, formatter=_default_flame_formatter, data_type='html', bind=False,
+                        on_next=False, on_next_args=()):
+    return {'formatter': formatter,
+            'data_type': data_type,
+            'bind': bind,
+            'on_next': on_next,
+            'on_next_args': on_next_args,
+            'order': max([c['order'] for c in existing_configs],
+                         default=-1) + 1,
+            }
+
+
+def flame(flame_key=None, formatter=_default_flame_formatter, data_type='html', bind=False, on_next=False,
+          on_next_args=(), decorator_name='flame'):
+    def decorator(func):
+        if type(func) is PromiseProxy:
+            raise Exception(f"@{decorator_name} must be applied to a function (after @app.task) on {func.__name__}")
+
+        undecorated = undecorate_func(func)
+        if not hasattr(undecorated, 'flame_data_configs'):
+            undecorated.flame_data_configs = OrderedDict()
+        undecorated.flame_data_configs[flame_key] = create_flame_config(undecorated.flame_data_configs.values(),
+                                                                        formatter, data_type, bind, on_next,
+                                                                        on_next_args)
+        return func
+
+    return decorator
+
+
 class FireXTask(Task):
     """
     Task object that facilitates passing of arguments and return values from one task to another, to be used in chains
@@ -92,7 +194,8 @@ class FireXTask(Task):
             self.initialize_context()
             yield
         finally:
-            del self.context
+            if hasattr(self, 'context'):
+                del self.context
 
     @property
     def from_plugin(self):
@@ -123,6 +226,12 @@ class FireXTask(Task):
         self.context.enqueued_children = {}
         self.context.bog = None
 
+        # Favour @app.task(flame=...) driven-style when present over @flame() annotation style.
+        # Flame configs need to be on self.context b/c they write to flame_data_configs[k]['on_next'] for collapse ops.
+        # Might make more sense to rework that to avoid flame data on context.
+        self.context.flame_configs = self.get_task_flame_configs() or deepcopy(getattr(self.undecorated,
+                                                                                       "flame_data_configs", {}))
+
     def get_module_file_location(self):
         return sys.modules[self.__module__].__file__
 
@@ -225,11 +334,13 @@ class FireXTask(Task):
                             firex_default_bound_args=convert_to_serializable(default_bound_args),
                             called_as_orig=self.called_as_orig,
                             long_name=self.name_without_orig,
-                            log_filepath=self.task_logfile,
+                            log_filepath=self.task_log_url,
                             from_plugin=self.from_plugin,
                             code_filepath=self.code_filepath,
                             retries=self.request.retries,
+                            task_parent_id=self.request.parent_id,
                             **extra_events)
+            self.send_firex_data(self.bag)
 
         # Print the pre-call header
         self.print_precall_header(bound_args, default_bound_args)
@@ -269,6 +380,7 @@ class FireXTask(Task):
         # Send a custom task-succeeded event with the results
         if not self.request.called_directly:
             self.send_event('task-results', firex_result=convert_to_serializable(results), **extra_events)
+            self.send_firex_data(self.bag)
 
     def print_precall_header(self, bound_args, default_bound_args):
         n = 1
@@ -335,7 +447,11 @@ class FireXTask(Task):
             finally:
                 self.remove_task_logfile_handler()
 
-    def handle_exception(self, e, logging_extra=None, raise_exception=True):
+    def handle_exception(self, e, logging_extra: dict=None, raise_exception=True):
+        extra = {'span_class': 'exception'}
+        if logging_extra:
+            extra.update(logging_extra)
+
         if isinstance(e, ChainInterruptedException) or isinstance(e, ChainRevokedException):
             try:
                 exception_cause_uuid = e.task_id
@@ -348,7 +464,7 @@ class FireXTask(Task):
         exception_string = str(e)
         if exception_string:
             mssg += f': {exception_string}'
-        logger.error(mssg, exc_info=e, extra=logging_extra)
+        logger.error(mssg, exc_info=e, extra=extra)
         if raise_exception:
             raise e
 
@@ -504,6 +620,8 @@ class FireXTask(Task):
 
     def wait_for_specific_children(self, child_results, **kwargs: dict):
         """Wait for the explicitly provided child_results to run and complete"""
+        if isinstance(child_results, AsyncResult):
+            child_results = [child_results]
         if child_results:
             logger.debug('Waiting for enqueued children: %r' % get_tasks_names_from_results(child_results))
             try:
@@ -547,6 +665,145 @@ class FireXTask(Task):
                     self._update_child_state(child_result, self._UNBLOCKED)
         return child_result
 
+    def enqueue_child_and_get_results(self,
+                                      *args,
+                                      return_keys: Union[str, tuple] = (),
+                                      return_keys_only: bool = True,
+                                      merge_children_results: bool = False,
+                                      extract_from_parents: bool = True,
+                                      **kwargs) -> Union[tuple, dict]:
+        """Apply a ``chain``, and extract results from it.
+
+        This is a better version of `enqueue_child_and_extract` where the defaults for
+        `extract_from_children` and `extract_task_returns_only` defaults are more intuitive.
+        Additionally, extract_from_parents defaults to True in this API.
+
+        Note:
+            This is shorthand for :meth:`enqueue_child` followed with :meth:`get_results`.
+
+        Args:
+            *args: Tuple of args required by  :meth:`enqueue_child`
+            return_keys: A single return key string, or a tuple of keys to extract from the task results.
+                The default value of :const:`None` will return a dictionary of key/value pairs for the returned results.
+            return_keys_only: If set, only return results for keys specified by the tasks' `@returns`
+                decorator or :attr:`returns` attribute, otherwise, returns will include key/value pairs from the BoG.
+            merge_children_results: If set, extract and merge results from the children tasks as well.
+            extract_from_parents: If set, will consider all results returned from tasks of the given chain (parents
+                of the last task). Else will consider only results returned by the last task of the chain.
+            **kwargs: Other options to :meth:`enqueue_child`
+
+        Returns:
+            The returns of `get_results`.
+
+        See Also:
+            get_results
+        """
+
+        return self.enqueue_child_and_extract(*args,
+                                              return_keys=return_keys,
+                                              extract_from_children=merge_children_results,
+                                              extract_task_returns_only=return_keys_only,
+                                              extract_from_parents=extract_from_parents,
+                                              **kwargs)
+
+    def enqueue_child_and_extract(self,
+                                  *args,
+                                  return_keys: Union[str, tuple] = (),
+                                  extract_from_children: bool = True,
+                                  extract_task_returns_only: bool = False,
+                                  extract_from_parents: bool = False,
+                                  **kwargs) -> Union[tuple, dict]:
+        """Apply a ``chain``, and extract results from it.
+
+        Note:
+            This is shorthand for :meth:`enqueue_child` followed with :meth:`extract_and_filter`.
+
+        Args:
+            *args: Tuple of args required by  :meth:`enqueue_child`
+            return_keys: A single return key string, or a tuple of keys to extract from the task results.
+                The default value of :const:`None` will return a dictionary of key/value pairs for the returned results.
+            extract_from_children: If set, extract and merge results from the children tasks as well.
+            extract_task_returns_only: If set, only return results for keys specified by the tasks' `@returns`
+                decorator or :attr:`returns` attribute, otherwise, returns will include key/value pairs from the BoG.
+            extract_from_parents: If set, will consider all results returned from tasks of the given chain (parents
+                of the last task). Else will consider only results returned by the last task of the chain.
+            **kwargs: Other options to :meth:`enqueue_child`
+
+        Returns:
+            The returns of `extract_and_filter`.
+
+        See Also:
+            extract_and_filter
+            enqueue_child_and_get_results
+        """
+
+        if kwargs.pop('enqueue_once_key', None):
+            raise ValueError('Invalid argument. Use the enqueue_child_once_and_extract() api.')
+
+        return self._enqueue_child_and_extract(*args,
+                                               return_keys=return_keys,
+                                               extract_from_children=extract_from_children,
+                                               extract_task_returns_only=extract_task_returns_only,
+                                               enqueue_once_key='',
+                                               extract_from_parents=extract_from_parents,
+                                               **kwargs)
+
+    def _enqueue_child_and_extract(self,
+                                   *args,
+                                   return_keys: Union[str, tuple] = (),
+                                   extract_from_children: bool = True,
+                                   extract_task_returns_only: bool = False,
+                                   enqueue_once_key: str = '',
+                                   extract_from_parents: bool = False,
+                                   **kwargs) -> Union[tuple, dict]:
+        """Apply a ``chain``, and extract results from it.
+
+        Note:
+            This is shorthand for :meth:`enqueue_child` followed with :meth:`extract_and_filter`.
+
+        Args:
+            *args: Tuple of args required by  :meth:`enqueue_child`
+            return_keys: A single return key string, or a tuple of keys to extract from the task results.
+                The default value of :const:`None` will return a dictionary of key/value pairs for the returned results.
+            extract_from_children: If set, extract and merge results from the children tasks as well.
+            extract_task_returns_only: If set, only return results for keys specified by the tasks' `@returns`
+                decorator or :attr:`returns` attribute, otherwise, returns will include key/value pairs from the BoG.
+            enqueue_once_key: a string key, which, if set will be used to check if this task needs to be run,
+                Use the enqueue_child_once_and_extract wrapper to set this. Should be unique per FireX run.
+            extract_from_parents: If set, will consider all results returned from tasks of the given chain (parents
+                of the last task). Else will consider only results returned by the last task of the chain.
+                NOTE: Will not work on reconstituted AsyncResult objects, such as those sometimes created by
+                the enqueue_once API.
+            **kwargs: Other options to :meth:`enqueue_child`
+
+        Returns:
+            The returns of `extract_and_filter`.
+
+        See Also:
+            extract_and_filter
+        """
+
+        # Remove block from kwargs if it exists
+        _block = kwargs.pop('block', True)
+        if not _block:
+            logger.warning(f'enqueue_child_and_extract ignored block={_block}, '
+                           'since it needs to block in order to extract results')
+
+        if not enqueue_once_key:
+            result_promise = self.enqueue_child(*args, block=True, **kwargs)
+        else:
+            # Need to make sure task with this key is run only once
+            result_promise = self._enqueue_child_once(*args,
+                                                      enqueue_once_key=enqueue_once_key,
+                                                      block=True,
+                                                      **kwargs)
+
+        return extract_and_filter(result_promise,
+                                  return_keys,
+                                  extract_from_children=extract_from_children,
+                                  extract_task_returns_only=extract_task_returns_only,
+                                  extract_from_parents=extract_from_parents)
+
     def enqueue_in_parallel(self, chains, max_parallel_chains=15, wait_for_completion=True,
                             raise_exception_on_failure=False):
         """ This method executes the provided list of Signatures/Chains in parallel
@@ -615,6 +872,15 @@ class FireXTask(Task):
     def get_task_logging_dirpath(file_logging_dirpath, hostname):
         return os.path.join(file_logging_dirpath, hostname)
 
+    @property
+    def task_log_url(self):
+        if self.app.conf.install_config.has_viewer():
+            # FIXME: there must be a more direct way of getting this relative path.
+            log_entry_rel_run_root = os.path.relpath(self.task_logfile, self.app.conf.logs_dir)
+            return self.app.conf.install_config.get_log_entry_url(log_entry_rel_run_root)
+        else:
+            return self.task_logfile
+
     @property
     def task_logfile(self):
         return self.get_task_logfile(self.task_logging_dirpath, self.name, self.request.id)
@@ -685,6 +951,85 @@ class FireXTask(Task):
     def start_time(self):
         return get_task_start_time(self.request.id, self.backend)
 
+    def get_task_flame_configs(self) -> OrderedDict:
+        flame_value = get_attr_unwrapped(self, 'flame', None)
+        task_flame_config = OrderedDict()
+        if flame_value:
+            if isinstance(flame_value, str):
+                # Config is only a key name, fill in default config.
+                task_flame_config = OrderedDict([(flame_value, create_flame_config([]))])
+            elif isinstance(flame_value, list):
+                # Config is a list of key names, each of which should get a default config.
+                # Create list of default configs so that their order is set properly.
+                default_flame_configs = []
+                for _ in flame_value:
+                    default_flame_configs.append(create_flame_config(default_flame_configs))
+                # associated ordered default configs with key names from flame decorator.
+                task_flame_config = OrderedDict([(key_name, default_flame_configs[i])
+                                                 for i, key_name in enumerate(flame_value)])
+            elif isinstance(flame_value, dict):
+                task_flame_config = OrderedDict(flame_value)
+
+        return task_flame_config
+
+    def send_firex_data(self, data):
+        if self.request.called_directly:
+            return
+
+        if getattr(self.context, 'flame_configs', False):
+            def safe_format(formatter, fromatter_args, formatter_kwargs):
+                try:
+                    return formatter(*fromatter_args, **formatter_kwargs)
+                except Exception as e:
+                    logger.error(e)
+                    return None
+
+            formatted_data = {}
+            for flame_key, flame_config in self.context.flame_configs.items():
+                # Data can be sent either because it is supplied in the input or if data was registered to be sent
+                # 'on_next' during flame_data_config registration.
+                if flame_key in data \
+                        or flame_config['on_next'] \
+                        or flame_key is None:
+                    formatter_kwargs = {'task': self} if flame_config['bind'] else {}
+                    if flame_key in data:
+                        formatter_args = [data[flame_key]]
+                    elif flame_config['on_next']:
+                        formatter_args = flame_config['on_next_args']
+                    elif flame_key is None:
+                        # None means execute formatter with all data.
+                        formatter_args = [data]
+                    else:
+                        formatter_args = []
+
+                    format_result = safe_format(flame_config['formatter'], formatter_args, formatter_kwargs)
+                    if format_result is not None:
+                        formatted_data[flame_key] = {
+                            'value': format_result,
+                            'type': flame_config['data_type'],
+                            'order': flame_config['order'],
+                        }
+
+            if formatted_data:
+                self.send_firex_event_raw({'flame_data': formatted_data})
+                sent_on_next_keys = [k for k, v in self.context.flame_configs.items()
+                                     if v['on_next'] and k in formatted_data]
+                for k in sent_on_next_keys:
+                    self.context.flame_configs[k]['on_next'] = False
+
+    def send_firex_event_raw(self, data):
+        self.send_event('task-send-flame', **data)
+
+    def update_firex_data(self, **kwargs):
+        self.send_firex_data(kwargs)
+
+    def send_firex_html(self, **kwargs):
+        formatted_data = {flame_key: {'value': html_data,
+                                      'type': 'html',
+                                      'order': time.time()}
+                          for flame_key, html_data in kwargs.items()}
+        self.send_firex_event_raw({'flame_data': formatted_data})
+
 
 def undecorate_func(func):
     undecorated_func = func
